import uuid
import datetime
from decimal import Decimal
import numpy as np


DB_TYPES: dict = {
    bool: "boolean",
    int: "integer",
    np.int64: "bigint",
    float: "float",
    str: "character varying",
    bytes: "byte",
    list: "Array",
    Decimal: "numeric",
    datetime.date: "date",
    datetime.datetime: "timestamp without time zone",
    datetime.time: "time",
    datetime.timedelta: "timestamp without time zone",
    uuid.UUID: "uuid",
    dict: "jsonb"
}

MODEL_TYPES = {
    "boolean": bool,
    "integer": int,
    "bigint": np.int64,
    "float": float,
    "character varying": str,
    "string": str,
    "varchar": str,
    "byte": bytes,
    "bytea": bytes,
    "Array": list,
    "hstore": dict,
    "character varying[]": list,
    "numeric": Decimal,
    "date": datetime.date,
    "timestamp with time zone": datetime.datetime,
    "time": datetime.time,
    "timestamp without time zone": datetime.datetime,
    "uuid": uuid.UUID,
    "json": dict,
    "jsonb": dict,
    "text": str,
    "serial": int,
    "bigserial": int,
    "inet": str,
}

JSON_TYPES = {
    bool: "boolean",
    int: "integer",
    np.int64: "integer",
    float: "float",
    str: "string",
    bytes: "byte",
    list: "list",
    Decimal: "decimal",
    datetime.date: "date",
    datetime.datetime: "datetime",
    datetime.time: "time",
    datetime.timedelta: "timedelta",
    uuid.UUID: "uuid",
}



"""AsyncDB Meta information."""

__title__ = 'python-datamodel'
__description__ = ('simple library based on python +3.7 to use Dataclass-syntax'
                   'for interacting with Data')
__version__ = '0.0.18'
__author__ = 'Jesus Lara'
__author_email__ = 'jesuslarag@gmail.com'
__license__ = 'BSD'



# -*- coding: utf-8 -*-
"""DataModels.

DataModel is a reimplementation of dataclasses with true inheritance and composition.
"""
from datamodel.fields import Field, Column
from .base import BaseModel
from .version import (
    __title__, __description__, __version__, __author__, __author_email__
)

__all__ = ('Field', 'Column', 'BaseModel', )



import uuid
from datamodel import BaseModel, Column
from datetime import datetime


def auto_now_add(*args, **kwargs):
    return uuid.uuid4()

class Contact(BaseModel):
    account: str = ''
    value: str = ''

class User(BaseModel):
    """
    User Basic Structure
    """
    id: uuid.UUID = Column(primary_key=True, required=True, default=auto_now_add(), db_default='uuid_generate_v4()')
    firstname: str
    lastname: str
    name: str = Column(required=True, default='John Doe')
    age: int = Column(default=18, required=True)
    signup_ts: datetime = Column(default=datetime.now(), db_default='now()')
    contacts: Contact = Column(required=False)

    class Meta:
        name = 'users'
        schema = 'public'
        driver = 'pg'
        strict = False
        michiko = 'mamon'

u = User()
print(u.model(dialect='json'))



"""
JSON Encoders.
"""
from dataclasses import _MISSING_TYPE, MISSING
import decimal
from typing import Any, Union
from decimal import Decimal
import asyncpg
import numpy as np
import orjson


class DefaultEncoder:
    """
    Basic Encoder using orjson
    """
    def __init__(self, **kwargs):
        # eventually take into consideration when serializing
        self.options = kwargs

    def __call__(self, obj, **kwargs) -> Any:
        return self.encode(obj, **kwargs)

    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(obj, Decimal):
            return str(obj)
        elif hasattr(obj, "isoformat"):
            return obj.isoformat()
        elif isinstance(obj, asyncpg.Range):
            return [obj.lower, obj.upper]
        elif hasattr(obj, "hex"):
            return obj.hex
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif obj == _MISSING_TYPE:
            return None
        elif obj == MISSING:
            return None
        raise TypeError(f"{obj!r} is not JSON serializable")

    def encode(self, obj, **kwargs) -> str:
        # decode back to str, as orjson returns bytes
        options = {
            "default": self.default,
            "option": orjson.OPT_NAIVE_UTC | orjson.OPT_SERIALIZE_NUMPY| orjson.OPT_UTC_Z
        }
        if kwargs:
            options = {**options, **kwargs}
        return orjson.dumps(
            obj,
            **options
        ).decode('utf-8')

    dumps = encode

    def decode(self, obj) -> Union[dict, list]:
        return orjson.loads(
            obj
        )

    loads = decode

class BaseEncoder:
    """
    Encoder replacement for json.dumps using orjson
    """

    def __init__(self, *args, **kwargs):
        encoder = DefaultEncoder(*args, **kwargs)
        self.encode = encoder.__call__



import uuid
from typing import Union, List, Optional
from dataclasses import dataclass, fields, is_dataclass, field
import orjson
from datamodel import Field, BaseModel


def auto_uid():
    return uuid.uuid4()

@dataclass
class Point:
    x: int = Field(default=0, min=0, max=10)
    y: int = Field(default=0, min=0, max=10)
    c: float = Field(default=10, init=False)


a = Point(x=10, y=10)
print(a)
print(fields(a))
print('IS a Dataclass?: ', is_dataclass(a))


class intSum(object):
    def get_coordinate(self):
        return (self.x, self.y)

# or using Model:
class newPoint(BaseModel, intSum):
    x: int = Field(default=0, min=0, max=10)
    y: Union[int, None] = Field(default=0, min=0, max=10)

a = newPoint(x=10, y=10)
print(a)
print(fields(a))
print('IS a Dataclass?: ', is_dataclass(a))
print(a.get_coordinate())

class coordinate(BaseModel, intSum):
    latitude: float
    longitude: float

    def get_location(self) -> tuple:
        return (self.latitude, self.longitude)


class Country(BaseModel):
    country: str = Field(required=True)
    code: str = Field(min=0, max=2)

def default_number():
    return '6'

def default_rect():
    return [0,0,0,0]

def valid_zipcode(field, value):
    return value == 45510

class Address(BaseModel):
    id: uuid.UUID = field(default_factory=auto_uid)
    street: str = Field(required=True)
    number: str = Field(factory=default_number)
    zipcode: int = Field(required=False, default=1010, validator=valid_zipcode)
    location: Optional[coordinate]
    country: Union[Country, None] = Field(required=False)
    box: List[Optional[newPoint]]
    rect: List[int] = Field(factory=default_rect)
    prueba: str = Field(required=False)

addr = Address(street="Beato Juan de Avila", location=(18.1, 22.1), zipcode=45510, box=[(2, 10), (4, 8)], rect=[1, 2, 3, 4])
print(addr)
print('IS a Dataclass?: ', is_dataclass(addr))

print(addr.location.get_location())
print('== Export to JSON ==')
print(addr.json())
b = addr.json()
data = orjson.loads(b)
data['country'] = {
    "country": "Spain",
    "code": "ES"
}
addr2 = Address(**data)
print(addr2)
print('== Using "from-json" method ==')
addr3 = Address.from_json(b)
print(addr3)
print('=== PRINTING MODEL === ')
print(Address.model())



import uuid
from typing import Union, List, Optional
from dataclasses import dataclass, field, is_dataclass
from datamodel import BaseModel, Column, Field

def auto_uid():
    return uuid.uuid4()

class User(BaseModel):
    name: str
    first_name: str
    last_name: str

@dataclass
class Address:
    street: str
    city: str
    state: str
    zipcode: str
    country: Optional[str] = 'US'

    def __str__(self) -> str:
        """Provides pretty response of address"""
        lines = [self.street]
        lines.append(f"{self.city}, {self.zipcode} {self.state}")
        lines.append(f"{self.country}")
        return "\n".join(lines)

class Employee(User):
    """
    Base Employee.
    """
    id: uuid.UUID = field(default_factory=auto_uid)
    role: str
    address: Address # composition of a dataclass inside of DataModel

    def info(self) -> str:
        return f"Name is {self.first_name}, {self.last_name}"

# Supporting multiple inheritance, even from pure dataclasses
# Wage Policies
class MonthlySalary(BaseModel):
    salary: Union[float, int]

    def calculate_payroll(self) -> Union[float, int]:
        return self.salary

class HourlySalary(BaseModel):
    salary: Union[float, int] = Field(default=0)
    hours_worked: Union[float, int] = Field(default=0)

    def calculate_payroll(self) -> Union[float, int]:
        return (self.hours_worked * self.salary)

class CommissionPolicy(BaseModel):
    commission: Union[int, float] = Column(default=0)

    def calculate_payroll(self) -> Union[float, int]:
        # adding the commission to the base payroll
        fixed = super().calculate_payroll()
        return fixed + self.commission

# TYpes of Employees
class Manager(Employee, MonthlySalary, CommissionPolicy):
    """
    Manager is an employee montly salary policy and commissions.
    """
    role: str = 'Manager'

class SalesPerson(Employee, HourlySalary, CommissionPolicy):
    """
    SalesPerson is an employee with hourly salary policy and commissions.
    """
    role: str = 'Sales Person'

class Secretary(Employee, MonthlySalary):
    """Secretary.

    Person with montly salary policy and no commissions.
    """
    role: str = 'Secretary'

class FactoryWorker(Employee, HourlySalary):
    """
    FactoryWorker is an employee with hourly salary policy and no commissions.
    """
    role: str = 'Factory Worker'

class PayrollSystem:
    def calculate_payroll(self, employees: List[dataclass]) -> None:
        print('=== Calculating Payroll === ')
        for employee in employees:
            print(f"Payroll for employee {employee.id} - {employee.name}")
            print(f"- {employee.role} Amount: {employee.calculate_payroll()}")
            if employee.address:
                print('- Sent to:')
                print(employee.address)
            print("")


# Testing Model:
john = Manager(name='John Doe', first_name='John', last_name='Doe', salary=2500, commission=500)
john.address = Address('121 Admin Road', "Concord", "NH", "03301")
jane = Secretary(name='Jane Doe', first_name='Jane', last_name='Doe', salary=1500)
jane.address = Address('Rodeo Drive, Rd', 'Los Angeles', 'CA', '31050')
bob = FactoryWorker(name='Bob Doyle', first_name='Bob', last_name='Doyle', salary=15, hours_worked=40)
mitch = FactoryWorker(name='Mitch Brian', first_name='Mitch', last_name='Brian', salary=20, hours_worked=35)
kevin = SalesPerson(name='Kevin Bacon', first_name='Kevin', last_name='Bacon', salary=35, hours_worked=35, commission=250)

payroll = PayrollSystem()
payroll.calculate_payroll([jane, bob, mitch, john, kevin])

print('IS DATACLASS? > ', is_dataclass(mitch))



from __future__ import annotations
import inspect
import types
from typing import (
    Optional,
    Union,
    Any
)
import logging
import uuid
from collections.abc import Callable
# Dataclass
from dataclasses import (
    dataclass,
    is_dataclass,
    _FIELD,
    asdict,
    make_dataclass,
    _MISSING_TYPE
)
from orjson import OPT_INDENT_2
from datamodel.fields import Field
from .parsers import DefaultEncoder
from .exceptions import (
    ValidationModel
)
from .types import JSON_TYPES

class Meta:
    """
    Metadata information about Model.
    """
    name: str = ""
    schema: str = ""
    app_label: str = ""
    frozen: bool = False
    strict: bool = True
    driver: str = None
    credentials: dict = Optional[dict]
    dsn: Optional[str] = None
    datasource: Optional[str] = None
    connection: Optional[Callable] = None

def set_connection(cls, conn: Callable):
    cls.connection = conn


def _dc_method_setattr(
            self,
            name: str,
            value: Any,
        ) -> None:
    """
    _dc_method_setattr.
    Method for overwrite the "setattr" on Dataclasses.
    """
    if self.Meta.frozen is True and name not in self.__fields__:
        raise TypeError(
            f"Cannot add New attribute {name} on {self.modelName}, "
            "This DataClass is frozen (read-only class)"
        )
    else:
        object.__setattr__(self, name, value)
        if name not in self.__fields__:
            if self.Meta.strict is True:
                logging.warning(
                    f"Warning: *{name}* doesn't exists on {self.modelName}"
                )
            else:
                try:
                    # create a new Field on Model.
                    f = Field(required=False, default=value)
                    f.name = name
                    f.type = type(value)
                    self.__columns__[name] = f
                    self.__fields__.append(name)
                    setattr(self, name, value)
                except Exception as err:
                    logging.exception(err, stack_info=True)
                    raise


def create_dataclass(
    new_cls: Union[object, Any],
    frozen: bool = False
) -> Callable:
    """
    create_dataclass.
       Create a Dataclass from a simple Class
    """
    dc = dataclass(unsafe_hash=True, init=True, frozen=frozen)(new_cls)
    setattr(dc, "__setattr__", _dc_method_setattr)
    # adding a properly internal json encoder:
    dc.__encoder__ = DefaultEncoder()
    dc.__valid__: bool = False
    return dc


class ModelMeta(type):
    """
    ModelMeta.
      MetaClass object to create dataclasses for modeling DataModels.
    """
    __fields__: list

    Meta = Meta

    def __new__(cls, name, bases, attrs, **kwargs):
        """__new__ is a classmethod, even without @classmethod decorator"""
        cols = []
        if "__annotations__" in attrs:
            annotations = attrs["__annotations__"]
            for field, _type in annotations.items():
                # print(f"Field: {field}, Type: {_type}")
                if field in attrs:
                    df = attrs[field]
                    if isinstance(df, Field):
                        setattr(cls, field, df)
                    else:
                        df = Field(factory=_type, required=False, default=df)
                        df.name = field
                        df.type = _type
                        setattr(cls, field, df)
                else:
                    # add a new field, based on type
                    df = Field(factory=_type, required=False)
                    df.name = field
                    df.type = _type
                    setattr(cls, field, df)
                cols.append(field)
            # set the slots of this class
            cls.__slots__ = tuple(cols)
        attr_meta = attrs.pop("Meta", None)
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        new_cls.Meta = attr_meta or getattr(new_cls, "Meta", Meta)
        if not new_cls.Meta:
            new_cls.Meta = Meta
        new_cls.Meta.set_connection = types.MethodType(
            set_connection, new_cls.Meta
        )
        try:
            frozen = new_cls.Meta.frozen
        except AttributeError:
            new_cls.Meta.frozen = False
            frozen = False
        # mix values from Meta to an existing Meta Class
        new_cls.Meta.__annotations__ = Meta.__annotations__
        for key, _ in Meta.__annotations__.items():
            if not hasattr(new_cls.Meta, key):
                try:
                    setattr(new_cls.Meta, key, None)
                except AttributeError as e:
                    logging.warning(e)
        # adding a "class init method"
        try:
            new_cls.__model_init__(
                new_cls,
                name,
                attrs
            )
        except AttributeError:
            pass
        dc = create_dataclass(
            new_cls,
            frozen=frozen
        )
        cols = {
            k: v
            for k, v in dc.__dict__["__dataclass_fields__"].items()
            if v._field_type == _FIELD
        }
        dc.__columns__ = cols
        dc.__fields__ = tuple(cols.keys())
        return dc

    def __init__(cls, *args, **kwargs) -> None:
        cls.modelName = cls.__name__
        if cls.Meta.strict:
            cls.__frozen__ = cls.Meta.strict
        else:
            cls.__frozen__ = False
        # Initialized Data Model = True
        cls.__initialised__ = True
        super(ModelMeta, cls).__init__(*args, **kwargs)


class BaseModel(metaclass=ModelMeta):
    """
    Model.

    Base Model for all DataModels.
    """
    Meta = Meta

    def __unicode__(self):
        return str(__class__)

    def columns(self):
        return self.__columns__

    def get_fields(self):
        return self.__fields__

    def column(self, name):
        return self.__columns__[name]

    def dict(self):
        return asdict(self)

    def json(self, **kwargs):
        encoder = self.__encoder__
        if len(kwargs) > 0: # re-configure the encoder
            encoder = DefaultEncoder(**kwargs)
        return encoder(asdict(self))

    def is_valid(self):
        return bool(self.__valid__)

    def create_field(self, name: str, value: Any) -> None:
        """create_field.
        create a new Field on Model (when strict is False).
        Args:
            name (str): name of the field
            value (Any): value to be assigned.
        """
        f = Field(required=False, default=value)
        f.name = name
        f.type = type(value)
        self.__columns__[name] = f
        setattr(self, name, value)

    def set(self, name: str, value: Any) -> None:
        """set.
        Alias for Create Field.
        Args:
            name (str): name of the field
            value (Any): value to be assigned.
        """
        if name not in self.__columns__:
            self.create_field(name, value)
        else:
            setattr(self, name, value)

    def _parse_type(self, F, data) -> object:
        # TODO: migrate to cython, using Type
        _type = F.type
        if _type.__module__ == 'typing':
            args = None
            try:
                args = _type.__args__
            except AttributeError:
                pass
            if _type._name == 'Dict' and isinstance(data, dict):
                return {k: self._parse_type(F.type.__args__[1], v) for k, v in data.items()}
            elif _type._name == 'List' and isinstance(data, (list, tuple)):
                arg = args[0]
                if arg.__module__ == 'typing': # nested typing
                    try:
                        t = arg.__args__[0]
                        if is_dataclass(t):
                            result = []
                            for x in data:
                                if isinstance(x, dict):
                                    result.append(t(**x))
                                else:
                                    result.append(t(*x))
                            return result
                        else:
                            return data
                    except AttributeError:
                        return data # data -as is-
                elif is_dataclass(arg):
                    return [arg(*x) for x in data]
                else:
                    return data
            elif _type._name is None:
                if isinstance(_type.__origin__, type(Union)):
                    t = args[0]
                    if is_dataclass(t):
                        if isinstance(data, dict):
                            data = t(**data)
                        elif isinstance(data, (list, tuple)):
                            data = t(*data)
                        else:
                            data = None
                    # F.type = args[0]
                    return data
                else:
                    pass
        else:
            return data

    def __post_init__(self) -> None:
        """
         Post init method.
        Fill fields with function-factory or calling validations
        """
        # checking if an attribute is already a dataclass:
        for _, f in self.__columns__.items():
            value = getattr(self, f.name)
            key = f.name
            # print(f'FIELD {key} = {value}', 'TYPE : ', f.type)
            if is_dataclass(f.type): # is already a dataclass
                if isinstance(value, dict):
                    new_val = f.type(**value)
                    setattr(self, key, new_val)
            elif f.type.__module__ == 'typing':  # a typing extension
                new_val = self._parse_type(f, value)
                setattr(self, key, new_val)
            elif isinstance(value, list):
                try:
                    sub_type = f.type.__args__[0]
                    if is_dataclass(sub_type):
                        # for every item
                        items = []
                        for item in value:
                            try:
                                if isinstance(item, dict):
                                    items.append(sub_type(**item))
                                else:
                                    items.append(item)
                            except (TypeError, AttributeError):
                                continue
                        setattr(self, key, items)
                except AttributeError:
                    setattr(self, key, value)
            elif value is None:
                is_missing = isinstance(f.default, _MISSING_TYPE)
                setattr(self, key, f.default_factory if is_missing else f.default)
            elif f.type == uuid.UUID:
                # TODO: Automatic conversion from other types, like datatime, etc
                uid = None
                try:
                    uid = uuid.UUID(str(value))
                except ValueError as e:
                    print(e)
                setattr(self, key, uid if uid else f.default)
            else:
                continue
        try:
            self._validation()
        except RuntimeError as err:
            logging.exception(err)

    def is_callable(self, value) -> bool:
        is_missing = (value == _MISSING_TYPE)
        return callable(value) if not is_missing else False

    def _validation(self) -> None:
        """
        _validation.
        TODO: cover validations as length, not_null, required, max, min, etc
        """
        errors = {}
        for _, f in self.__columns__.items():
            name = f.name
            val = self.__dict__[name]
            if hasattr(f, 'default') and self.is_callable(val):
                try:
                    if val.__module__ != 'typing':
                        try:
                            new_val = val()
                        except TypeError:
                            try:
                                new_val = f.default()
                            except TypeError:
                                new_val = None
                        setattr(self, name, new_val)
                except TypeError as e:
                    print(e)
                    logging.warning(
                        f'{self.modelName}: Missing *Column* {f} with name {name}'
                    )
                    setattr(self, name, None)
            # first check: data type hint
            val = self.__dict__[name]
            val_type = type(val)
            annotated_type = f.type
            if val_type == type or val == annotated_type or val is None:
                # data not provided
                if f.metadata["required"] is True and self.Meta.strict is True:
                    errors[name] = ValidationModel(
                        field=name,
                        value=None,
                        value_type=val_type,
                        error="Field Required",
                        annotation=annotated_type,
                        exception=None,
                    )
                elif f.metadata['nullable'] is False:
                    errors[name] = ValidationModel(
                        field=name,
                        value=None,
                        value_type=val_type,
                        error="Not Null",
                        annotation=annotated_type,
                        exception=None,
                    )
            else:
                try:
                    instance = self._is_instanceof(val, annotated_type)
                    if not instance:
                            errors[name] = ValidationModel(
                                field=name,
                                value=val,
                                error="Validation Exception",
                                value_type=val_type,
                                annotation=annotated_type,
                                exception=None,
                            )
                except (TypeError) as e:
                    errors[name] = ValidationModel(
                        field=name,
                        value=val,
                        error="Validation Exception",
                        value_type=val_type,
                        annotation=annotated_type,
                        exception=e,
                    )
                ## calling validator:
                if 'validator' in f.metadata:
                    if f.metadata['validator'] is not None:
                        fn = f.metadata['validator']
                        if self.is_callable(fn):
                            try:
                                result = fn(f, val)
                                if result is False:
                                    errors[name] = ValidationModel(
                                        field=name,
                                        value=val,
                                        error=f"Validator: {result}",
                                        value_type=val_type,
                                        annotation=annotated_type,
                                        exception=None,
                                    )
                            except (ValueError, AttributeError, TypeError) as e:
                                errors[name] = ValidationModel(
                                    field=name,
                                    value=val,
                                    error="Validator Exception",
                                    value_type=val_type,
                                    annotation=annotated_type,
                                    exception=e,
                                )
        if errors:
            print("=== ERRORS ===")
            print(errors)
            object.__setattr__(self, "__valid__", False)
        else:
            object.__setattr__(self, "__valid__", True)

    def _is_instanceof(self, value: Any, annotated_type: type) -> bool:
        if annotated_type.__module__ == 'typing':
            return True # TODO: validate subscripted generic (typing extensions)
        else:
            try:
                return isinstance(value, annotated_type)
            except (AttributeError, TypeError, ValueError) as e:
                logging.error(e)
                raise

    @classmethod
    def make_model(cls, name: str, schema: str = "public", fields: list = None):
        parent = inspect.getmro(cls)
        obj = make_dataclass(name, fields, bases=(parent[0],))
        m = Meta()
        m.name = name
        m.schema = schema
        m.app_label = schema
        obj.Meta = m
        return obj

    @classmethod
    def from_json(cls, obj: str) -> dataclass:
        try:
            decoded = cls.__encoder__.loads(obj)
            return cls(**decoded)
        except ValueError as e:
            raise RuntimeError(
                "DataModel: Invalid string (JSON) data for decoding: {e}"
            ) from e

    @classmethod
    def from_dict(cls, obj: dict) -> dataclass:
        try:
            return cls(**obj)
        except ValueError as e:
            raise RuntimeError(
                "DataModel: Invalid Dictionary data for decoding: {e}"
            ) from e

    @classmethod
    def model(cls, dialect: str = "json") -> Any:
        """model.

        Return the json-version of current Model.
        Returns:
            str: string (json) version of model.
        """
        result = None
        clsname = cls.__name__
        schema = cls.Meta.schema
        table = cls.Meta.name if cls.Meta.name else clsname.lower()
        columns = cls.columns(cls).items()
        if dialect == 'json':
            cols = {}
            for _, field in columns:
                key = field.name
                _type = field.type
                if _type.__module__ == 'typing':
                    # TODO: discover real value of typing
                    if _type._name == 'List':
                        t = 'list'
                    elif _type._name == 'Dict':
                        t = 'dict'
                    else:
                        try:
                            t = _type.__args__[0]
                            t = t.__name__
                        except (AttributeError, ValueError):
                            t = 'object'
                else:
                    try:
                        t = JSON_TYPES[_type]
                    except KeyError:
                        t = 'object'
                cols[key] = {"name": key, "type": t}
            doc = {
                "name": clsname,
                "description": cls.__doc__.strip("\n").strip(),
                "table": table,
                "schema": schema,
                "fields": cols,
            }
            result = cls.__encoder__.dumps(doc, option=OPT_INDENT_2)
        return result



#!/usr/bin/env python
"""DataModels.

    Dataclass Reimplementation with true inheritance (without decorators.)
See:
https://github.com/phenobarbital/DataModel
"""
import sys
from os import path
from setuptools import find_packages, setup, Extension
from Cython.Build import cythonize


def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md'), encoding='utf-8') as rd:
        return rd.read()


with open(get_path('datamodel/version.py'), encoding='utf-8') as meta:
    exec(meta.read())

COMPILE_ARGS = ["-O2"]

extensions = [
    Extension(
        name='datamodel.fields',
        sources=['datamodel/fields.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    )
]

# sys.argv.extend(['--plat-name', 'manylinux_2_28_x86_64'])

setup(
    name="python-datamodel",
    version=__version__,
    python_requires=">=3.7.1",
    url="https://github.com/phenobarbital/python-datamodel",
    description=__description__,
    keywords=['asyncio', 'dataclass', 'dataclasses', 'data models'],
    platforms=['any'],
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        "Environment :: Web Environment",
    ],
    author=__author__,
    author_email=__author_email__,
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    include_package_data=True,
    license=__license__,
    setup_requires=[
        "setuptools==65.3.0",
        "wheel==0.37.1",
        "Cython==0.29.32",
        "asyncio==3.4.3",
        "cchardet==2.1.7"
    ],
    install_requires=[
        "wheel==0.37.1",
        "Cython==0.29.32",
        "numpy==1.23.3",
        "uvloop==0.17.0",
        "asyncio==3.4.3",
        "cchardet==2.1.7",
        "objectpath==0.6.1",
        "orjson==3.8.0",
        'typing_extensions==4.3.0',
        "asyncpg==0.26.0"
    ],
    tests_require=[
        'pytest>=6.0.0',
        'pytest-asyncio==0.19.0',
        'pytest-xdist==2.1.0',
        'pytest-assume==2.4.3'
    ],
    test_suite='tests',
    ext_modules=cythonize(extensions),
    project_urls={  # Optional
        "Source": "https://github.com/phenobarbital/datamodels",
        "Funding": "https://paypal.me/phenobarbital",
        "Say Thanks!": "https://saythanks.io/to/phenobarbital",
    },
)



"""
Base Exceptions for DataModel.
"""
from dataclasses import dataclass
from typing import (
    Optional,
    Union,
    Any
)


class ModelException(Exception):
    """Base class for other exceptions"""
    message: str = ''

    def __init__(self, message: str, *args: list) -> None:
        self.args = (
            message,
            *args
        )
        self.message = message
        super().__init__(message)

    def __repr__(self):
        return f"{__name__}({self.args!r})"

    def __str__(self):
        return f"{__name__}: {self.message}"


class ValidationError(ModelException):
    """Validation Error."""

    def __init__(self, message: str, *args: list) -> None:
        message = f'Validator Error: {message}'
        super().__init__(message)


class ParsingError(ModelException):
    """Parsing Error."""

    def __init__(self, message: str, *args: list) -> None:
        message = f'Parsing Error: {message}'
        super().__init__(message)


@dataclass
class ValidationModel:
    """
    Class for Error validation on DataModels.
    """
    field: str
    value: Optional[Union[str, Any]]
    error: str
    value_type: Any
    annotation: type
    exception: Optional[Exception]
